(* autogenerated from lockservice *)
From Perennial.goose_lang Require Import prelude.
From Perennial.program_proof.lockservice Require Import grove_prelude.

From Goose Require github_com.mit_pdos.go_journal.lockmap.
From Goose Require github_com.mit_pdos.lockservice.grove_common.
From Goose Require github_com.tchajed.marshal.

(* 0_common.go *)

Definition nondet: val :=
  rec: "nondet" <> :=
    #true.

(* Call this before doing an increment that has risk of overflowing.
   If it's going to overflow, this'll loop forever, so the bad addition can never happen *)
Definition overflow_guard_incr: val :=
  rec: "overflow_guard_incr" "v" :=
    Skip;;
    (for: (λ: <>, "v" + #1 < "v"); (λ: <>, Skip) := λ: <>,
      Continue).

(* 0_rpc.go *)

Definition RpcCoreHandler: ty := (struct.t grove_common.RPCVals.S -> uint64T)%ht.

Definition CheckReplyTable: val :=
  rec: "CheckReplyTable" "lastSeq" "lastReply" "CID" "Seq" "reply" :=
    let: ("last", "ok") := MapGet "lastSeq" "CID" in
    struct.storeF grove_common.RPCReply.S "Stale" "reply" #false;;
    (if: "ok" && ("Seq" ≤ "last")
    then
      (if: "Seq" < "last"
      then
        struct.storeF grove_common.RPCReply.S "Stale" "reply" #true;;
        #true
      else
        struct.storeF grove_common.RPCReply.S "Ret" "reply" (Fst (MapGet "lastReply" "CID"));;
        #true)
    else
      MapInsert "lastSeq" "CID" "Seq";;
      #false).

Definition rpcReqEncode: val :=
  rec: "rpcReqEncode" "req" :=
    let: "e" := marshal.NewEnc (#4 * #8) in
    marshal.Enc__PutInt "e" (struct.loadF grove_common.RPCRequest.S "CID" "req");;
    marshal.Enc__PutInt "e" (struct.loadF grove_common.RPCRequest.S "Seq" "req");;
    marshal.Enc__PutInt "e" (struct.get grove_common.RPCVals.S "U64_1" (struct.loadF grove_common.RPCRequest.S "Args" "req"));;
    marshal.Enc__PutInt "e" (struct.get grove_common.RPCVals.S "U64_2" (struct.loadF grove_common.RPCRequest.S "Args" "req"));;
    let: "res" := marshal.Enc__Finish "e" in
    "res".

Definition rpcReqDecode: val :=
  rec: "rpcReqDecode" "data" "req" :=
    let: "d" := marshal.NewDec "data" in
    struct.storeF grove_common.RPCRequest.S "CID" "req" (marshal.Dec__GetInt "d");;
    struct.storeF grove_common.RPCRequest.S "Seq" "req" (marshal.Dec__GetInt "d");;
    struct.storeF grove_common.RPCVals.S "U64_1" (struct.fieldRef grove_common.RPCRequest.S "Args" "req") (marshal.Dec__GetInt "d");;
    struct.storeF grove_common.RPCVals.S "U64_2" (struct.fieldRef grove_common.RPCRequest.S "Args" "req") (marshal.Dec__GetInt "d").

Definition rpcReplyEncode: val :=
  rec: "rpcReplyEncode" "reply" :=
    let: "e" := marshal.NewEnc (#2 * #8) in
    marshal.Enc__PutBool "e" (struct.loadF grove_common.RPCReply.S "Stale" "reply");;
    marshal.Enc__PutInt "e" (struct.loadF grove_common.RPCReply.S "Ret" "reply");;
    let: "res" := marshal.Enc__Finish "e" in
    "res".

Definition rpcReplyDecode: val :=
  rec: "rpcReplyDecode" "data" "reply" :=
    let: "d" := marshal.NewDec "data" in
    struct.storeF grove_common.RPCReply.S "Stale" "reply" (marshal.Dec__GetBool "d");;
    struct.storeF grove_common.RPCReply.S "Ret" "reply" (marshal.Dec__GetInt "d").

(* Essentially maps r -> e r e^{-1}, where e is the encode function and e^{-1} is decode
   function *)
Definition ConjugateRpcFunc: val :=
  rec: "ConjugateRpcFunc" "r" :=
    (λ: "rawReq" "rawRep",
      let: "req" := struct.alloc grove_common.RPCRequest.S (zero_val (struct.t grove_common.RPCRequest.S)) in
      let: "rep" := struct.alloc grove_common.RPCReply.S (zero_val (struct.t grove_common.RPCReply.S)) in
      rpcReqDecode "rawReq" "req";;
      "r" "req" "rep";;
      "rawRep" <-[slice.T byteT] rpcReplyEncode "rep";;
      #()
      ).

(* Common code for RPC clients: tracking of CID and next sequence number. *)
Module RPCClient.
  Definition S := struct.decl [
    "cid" :: uint64T;
    "seq" :: uint64T;
    "rawCl" :: struct.ptrT grove_ffi.RPCClient.S
  ].
End RPCClient.

Definition MakeRPCClient: val :=
  rec: "MakeRPCClient" "host" "cid" :=
    struct.new RPCClient.S [
      "cid" ::= "cid";
      "seq" ::= #1;
      "rawCl" ::= grove_ffi.MakeRPCClient "host"
    ].

(* 2 refers to the number of u64s in args *)
Definition RemoteProcedureCall2: val :=
  rec: "RemoteProcedureCall2" "cl" "rpcid" "req" "rep" :=
    let: "rawReq" := rpcReqEncode "req" in
    let: "rawRep" := ref (zero_val (slice.T byteT)) in
    let: "errb" := grove_ffi.RPCClient__RemoteProcedureCall "cl" "rpcid" "rawReq" "rawRep" in
    (if: ("errb" = #false)
    then
      rpcReplyDecode (![slice.T byteT] "rawRep") "rep";;
      #()
    else #());;
    "errb".

Definition RPCClient__MakeRequest: val :=
  rec: "RPCClient__MakeRequest" "cl" "rpcid" "args" :=
    overflow_guard_incr (struct.loadF RPCClient.S "seq" "cl");;
    let: "req" := struct.new grove_common.RPCRequest.S [
      "Args" ::= "args";
      "CID" ::= struct.loadF RPCClient.S "cid" "cl";
      "Seq" ::= struct.loadF RPCClient.S "seq" "cl"
    ] in
    let: "reply" := struct.alloc grove_common.RPCReply.S (zero_val (struct.t grove_common.RPCReply.S)) in
    struct.storeF RPCClient.S "seq" "cl" (struct.loadF RPCClient.S "seq" "cl" + #1);;
    let: "errb" := ref_to boolT #false in
    Skip;;
    (for: (λ: <>, #true); (λ: <>, Skip) := λ: <>,
      "errb" <-[boolT] RemoteProcedureCall2 (struct.loadF RPCClient.S "rawCl" "cl") "rpcid" "req" "reply";;
      (if: (![boolT] "errb" = #false)
      then Break
      else Continue));;
    struct.loadF grove_common.RPCReply.S "Ret" "reply".

(* Common code for RPC servers: handling of stale and redundant requests through
   the reply table. *)
Module RPCServer.
  Definition S := struct.decl [
    "lastSeq" :: mapT uint64T;
    "lastReply" :: mapT uint64T
  ].
End RPCServer.

Definition MakeRPCServer: val :=
  rec: "MakeRPCServer" <> :=
    let: "sv" := struct.alloc RPCServer.S (zero_val (struct.t RPCServer.S)) in
    struct.storeF RPCServer.S "lastSeq" "sv" (NewMap uint64T);;
    struct.storeF RPCServer.S "lastReply" "sv" (NewMap uint64T);;
    "sv".

Definition RPCServer__HandleRequest: val :=
  rec: "RPCServer__HandleRequest" "sv" "core" "req" "reply" :=
    (if: CheckReplyTable (struct.loadF RPCServer.S "lastSeq" "sv") (struct.loadF RPCServer.S "lastReply" "sv") (struct.loadF grove_common.RPCRequest.S "CID" "req") (struct.loadF grove_common.RPCRequest.S "Seq" "req") "reply"
    then #()
    else
      struct.storeF grove_common.RPCReply.S "Ret" "reply" ("core" (struct.loadF grove_common.RPCRequest.S "Args" "req"));;
      MapInsert (struct.loadF RPCServer.S "lastReply" "sv") (struct.loadF grove_common.RPCRequest.S "CID" "req") (struct.loadF grove_common.RPCReply.S "Ret" "reply"));;
    #().

(* 1_2pc.go *)

Module TxnResources.
  Definition S := struct.decl [
    "key" :: uint64T;
    "oldValue" :: uint64T
  ].
End TxnResources.

Module ParticipantServer.
  Definition S := struct.decl [
    "mu" :: lockRefT;
    "lockmap" :: struct.ptrT lockmap.LockMap.S;
    "kvs" :: mapT uint64T;
    "txns" :: mapT (struct.t TxnResources.S);
    "finishedTxns" :: mapT boolT
  ].
End ParticipantServer.

(* Precondition: emp
   returns 0 -> Vote Yes
   returns 1 -> Vote No *)
Definition ParticipantServer__PrepareIncrease: val :=
  rec: "ParticipantServer__PrepareIncrease" "ps" "tid" "key" "amount" :=
    lock.acquire (struct.loadF ParticipantServer.S "mu" "ps");;
    let: (<>, "ok") := MapGet (struct.loadF ParticipantServer.S "txns" "ps") "tid" in
    (if: "ok"
    then
      lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
      #0
    else
      let: (<>, "ok2") := MapGet (struct.loadF ParticipantServer.S "finishedTxns" "ps") "tid" in
      (if: "ok2"
      then
        lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
        #1
      else
        lockmap.LockMap__Acquire (struct.loadF ParticipantServer.S "lockmap" "ps") "key";;
        (if: "amount" + Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key") < Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key")
        then
          lockmap.LockMap__Release (struct.loadF ParticipantServer.S "lockmap" "ps") "key";;
          lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
          #1
        else
          MapInsert (struct.loadF ParticipantServer.S "txns" "ps") "tid" (struct.mk TxnResources.S [
            "key" ::= "key";
            "oldValue" ::= Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key")
          ]);;
          MapInsert (struct.loadF ParticipantServer.S "kvs" "ps") "key" (Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key") + "amount");;
          lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
          #0))).

Definition ParticipantServer__PrepareDecrease: val :=
  rec: "ParticipantServer__PrepareDecrease" "ps" "tid" "key" "amount" :=
    lock.acquire (struct.loadF ParticipantServer.S "mu" "ps");;
    let: (<>, "ok") := MapGet (struct.loadF ParticipantServer.S "txns" "ps") "tid" in
    (if: "ok"
    then
      lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
      #0
    else
      let: (<>, "ok2") := MapGet (struct.loadF ParticipantServer.S "finishedTxns" "ps") "tid" in
      (if: "ok2"
      then
        lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
        #1
      else
        lockmap.LockMap__Acquire (struct.loadF ParticipantServer.S "lockmap" "ps") "key";;
        (if: "amount" > Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key")
        then
          lockmap.LockMap__Release (struct.loadF ParticipantServer.S "lockmap" "ps") "key";;
          lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
          #1
        else
          MapInsert (struct.loadF ParticipantServer.S "txns" "ps") "tid" (struct.mk TxnResources.S [
            "key" ::= "key";
            "oldValue" ::= Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key")
          ]);;
          MapInsert (struct.loadF ParticipantServer.S "kvs" "ps") "key" (Fst (MapGet (struct.loadF ParticipantServer.S "kvs" "ps") "key") - "amount");;
          lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
          #0))).

Definition ParticipantServer__Commit: val :=
  rec: "ParticipantServer__Commit" "ps" "tid" :=
    lock.acquire (struct.loadF ParticipantServer.S "mu" "ps");;
    let: ("t", "ok") := MapGet (struct.loadF ParticipantServer.S "txns" "ps") "tid" in
    (if: ~ "ok"
    then
      lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
      #()
    else
      lockmap.LockMap__Release (struct.loadF ParticipantServer.S "lockmap" "ps") (struct.get TxnResources.S "key" "t");;
      MapDelete (struct.loadF ParticipantServer.S "txns" "ps") "tid";;
      MapInsert (struct.loadF ParticipantServer.S "finishedTxns" "ps") "tid" #true;;
      lock.release (struct.loadF ParticipantServer.S "mu" "ps")).

Definition ParticipantServer__Abort: val :=
  rec: "ParticipantServer__Abort" "ps" "tid" :=
    lock.acquire (struct.loadF ParticipantServer.S "mu" "ps");;
    let: ("t", "ok") := MapGet (struct.loadF ParticipantServer.S "txns" "ps") "tid" in
    (if: ~ "ok"
    then
      lock.release (struct.loadF ParticipantServer.S "mu" "ps");;
      #()
    else
      MapInsert (struct.loadF ParticipantServer.S "kvs" "ps") (struct.get TxnResources.S "key" "t") (struct.get TxnResources.S "oldValue" "t");;
      lockmap.LockMap__Release (struct.loadF ParticipantServer.S "lockmap" "ps") (struct.get TxnResources.S "key" "t");;
      MapDelete (struct.loadF ParticipantServer.S "txns" "ps") "tid";;
      MapInsert (struct.loadF ParticipantServer.S "finishedTxns" "ps") "tid" #true;;
      lock.release (struct.loadF ParticipantServer.S "mu" "ps")).

Definition MakeParticipantServer: val :=
  rec: "MakeParticipantServer" <> :=
    let: "s" := struct.alloc ParticipantServer.S (zero_val (struct.t ParticipantServer.S)) in
    struct.storeF ParticipantServer.S "mu" "s" (lock.new #());;
    struct.storeF ParticipantServer.S "lockmap" "s" (lockmap.MkLockMap #());;
    struct.storeF ParticipantServer.S "kvs" "s" (NewMap uint64T);;
    struct.storeF ParticipantServer.S "txns" "s" (NewMap (struct.t TxnResources.S));;
    struct.storeF ParticipantServer.S "finishedTxns" "s" (NewMap boolT).

Module TransactionCoordinator.
  Definition S := struct.decl [
    "s0" :: struct.ptrT ParticipantServer.S;
    "s1" :: struct.ptrT ParticipantServer.S
  ].
End TransactionCoordinator.

(* transfers between acc1 on s0 and acc2 on s1
   could also shard key-space *)
Definition TransactionCoordinator__doTransfer: val :=
  rec: "TransactionCoordinator__doTransfer" "tc" "tid" "acc1" "acc2" "amount" :=
    let: "prepared1" := ParticipantServer__PrepareIncrease (struct.loadF TransactionCoordinator.S "s0" "tc") "tid" "acc1" "amount" in
    let: "prepared2" := ParticipantServer__PrepareDecrease (struct.loadF TransactionCoordinator.S "s1" "tc") "tid" "acc2" "amount" in
    (if: ("prepared1" = #0) && ("prepared2" = #0)
    then
      ParticipantServer__Commit (struct.loadF TransactionCoordinator.S "s0" "tc") "tid";;
      ParticipantServer__Commit (struct.loadF TransactionCoordinator.S "s1" "tc") "tid"
    else
      ParticipantServer__Abort (struct.loadF TransactionCoordinator.S "s0" "tc") "tid";;
      ParticipantServer__Abort (struct.loadF TransactionCoordinator.S "s1" "tc") "tid").

(* 1_durable_kvserver.go *)

Module DurableKVServer.
  Definition S := struct.decl [
    "mu" :: lockRefT;
    "sv" :: struct.ptrT RPCServer.S;
    "kvs" :: mapT uint64T
  ].
End DurableKVServer.

Definition DurableKVServer__put_core: val :=
  rec: "DurableKVServer__put_core" "ks" "args" :=
    MapInsert (struct.loadF DurableKVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args") (struct.get grove_common.RPCVals.S "U64_2" "args");;
    #0.

Definition DurableKVServer__get_core: val :=
  rec: "DurableKVServer__get_core" "ks" "args" :=
    Fst (MapGet (struct.loadF DurableKVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args")).

Definition DurableKVServer__cas_core: val :=
  rec: "DurableKVServer__cas_core" "ks" "args" :=
    let: "r" := Fst (MapGet (struct.loadF DurableKVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args")) in
    (if: ("r" = struct.get grove_common.RPCVals.S "U64_2" "args")
    then
      MapInsert (struct.loadF DurableKVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args") (struct.get grove_common.RPCVals.S "U64_2" "args");;
      #1
    else #0).

(* requires (2n + 1) uint64s worth of space in the encoder *)
Definition EncMap: val :=
  rec: "EncMap" "e" "m" :=
    marshal.Enc__PutInt "e" (MapLen "m");;
    MapIter "m" (λ: "key" "value",
      marshal.Enc__PutInt "e" "key";;
      marshal.Enc__PutInt "e" "value").

Definition DecMap: val :=
  rec: "DecMap" "d" :=
    let: "sz" := marshal.Dec__GetInt "d" in
    let: "m" := NewMap uint64T in
    let: "i" := ref_to uint64T #0 in
    Skip;;
    (for: (λ: <>, ![uint64T] "i" < "sz"); (λ: <>, Skip) := λ: <>,
      let: "k" := marshal.Dec__GetInt "d" in
      let: "v" := marshal.Dec__GetInt "d" in
      MapInsert "m" "k" "v";;
      "i" <-[uint64T] ![uint64T] "i" + #1;;
      Continue);;
    "m".

(* For now, there is only one kv server in the whole world
   Assume it's in file "kvdur" *)
Definition WriteDurableKVServer: val :=
  rec: "WriteDurableKVServer" "ks" :=
    let: "num_bytes" := #8 * #2 * MapLen (struct.loadF RPCServer.S "lastSeq" (struct.loadF DurableKVServer.S "sv" "ks")) + #2 * MapLen (struct.loadF RPCServer.S "lastSeq" (struct.loadF DurableKVServer.S "sv" "ks")) + #2 * MapLen (struct.loadF DurableKVServer.S "kvs" "ks") + #3 in
    let: "e" := marshal.NewEnc "num_bytes" in
    EncMap "e" (struct.loadF RPCServer.S "lastSeq" (struct.loadF DurableKVServer.S "sv" "ks"));;
    EncMap "e" (struct.loadF RPCServer.S "lastReply" (struct.loadF DurableKVServer.S "sv" "ks"));;
    EncMap "e" (struct.loadF DurableKVServer.S "kvs" "ks");;
    grove_ffi.Write #(str"kvdur") (marshal.Enc__Finish "e");;
    #().

Definition ReadDurableKVServer: val :=
  rec: "ReadDurableKVServer" <> :=
    let: "content" := grove_ffi.Read #(str"kvdur") in
    (if: (slice.len "content" = #0)
    then slice.nil
    else
      let: "d" := marshal.NewDec "content" in
      let: "ks" := struct.alloc DurableKVServer.S (zero_val (struct.t DurableKVServer.S)) in
      let: "sv" := struct.alloc RPCServer.S (zero_val (struct.t RPCServer.S)) in
      struct.storeF RPCServer.S "lastSeq" "sv" (DecMap "d");;
      struct.storeF RPCServer.S "lastReply" "sv" (DecMap "d");;
      struct.storeF DurableKVServer.S "kvs" "ks" (DecMap "d");;
      struct.storeF DurableKVServer.S "sv" "ks" "sv";;
      struct.storeF DurableKVServer.S "mu" "ks" (lock.new #());;
      "ks").

Definition DurableKVServer__Put: val :=
  rec: "DurableKVServer__Put" "ks" "req" "reply" :=
    lock.acquire (struct.loadF DurableKVServer.S "mu" "ks");;
    RPCServer__HandleRequest (struct.loadF DurableKVServer.S "sv" "ks") (DurableKVServer__put_core "ks") "req" "reply";;
    WriteDurableKVServer "ks";;
    lock.release (struct.loadF DurableKVServer.S "mu" "ks");;
    #().

Definition DurableKVServer__Get: val :=
  rec: "DurableKVServer__Get" "ks" "req" "reply" :=
    lock.acquire (struct.loadF DurableKVServer.S "mu" "ks");;
    RPCServer__HandleRequest (struct.loadF DurableKVServer.S "sv" "ks") (DurableKVServer__get_core "ks") "req" "reply";;
    WriteDurableKVServer "ks";;
    lock.release (struct.loadF DurableKVServer.S "mu" "ks");;
    #().

Definition DurableKVServer__CAS: val :=
  rec: "DurableKVServer__CAS" "ks" "req" "reply" :=
    lock.acquire (struct.loadF DurableKVServer.S "mu" "ks");;
    RPCServer__HandleRequest (struct.loadF DurableKVServer.S "sv" "ks") (DurableKVServer__cas_core "ks") "req" "reply";;
    WriteDurableKVServer "ks";;
    lock.release (struct.loadF DurableKVServer.S "mu" "ks");;
    #().

Definition MakeDurableKVServer: val :=
  rec: "MakeDurableKVServer" <> :=
    let: "ks_old" := ReadDurableKVServer #() in
    (if: "ks_old" ≠ #null
    then "ks_old"
    else
      let: "ks" := struct.alloc DurableKVServer.S (zero_val (struct.t DurableKVServer.S)) in
      struct.storeF DurableKVServer.S "mu" "ks" (lock.new #());;
      struct.storeF DurableKVServer.S "kvs" "ks" (NewMap uint64T);;
      struct.storeF DurableKVServer.S "sv" "ks" (MakeRPCServer #());;
      "ks").

(* Does not return *)
Definition DurableKVServer__Start: val :=
  rec: "DurableKVServer__Start" "ks" :=
    let: "handlers" := NewMap grove_common.RawRpcFunc in
    grove_ffi.StartRPCServer "handlers".

(* 1_kvserver.go *)

Definition KV_PUT : expr := #1.

Definition KV_GET : expr := #2.

Module KVServer.
  Definition S := struct.decl [
    "mu" :: lockRefT;
    "sv" :: struct.ptrT RPCServer.S;
    "kvs" :: mapT uint64T
  ].
End KVServer.

Definition KVServer__put_core: val :=
  rec: "KVServer__put_core" "ks" "args" :=
    MapInsert (struct.loadF KVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args") (struct.get grove_common.RPCVals.S "U64_2" "args");;
    #0.

Definition KVServer__get_core: val :=
  rec: "KVServer__get_core" "ks" "args" :=
    Fst (MapGet (struct.loadF KVServer.S "kvs" "ks") (struct.get grove_common.RPCVals.S "U64_1" "args")).

Definition KVServer__Put: val :=
  rec: "KVServer__Put" "ks" "req" "reply" :=
    lock.acquire (struct.loadF KVServer.S "mu" "ks");;
    RPCServer__HandleRequest (struct.loadF KVServer.S "sv" "ks") (KVServer__put_core "ks") "req" "reply";;
    lock.release (struct.loadF KVServer.S "mu" "ks");;
    #().

Definition KVServer__Get: val :=
  rec: "KVServer__Get" "ks" "req" "reply" :=
    lock.acquire (struct.loadF KVServer.S "mu" "ks");;
    RPCServer__HandleRequest (struct.loadF KVServer.S "sv" "ks") (KVServer__get_core "ks") "req" "reply";;
    lock.release (struct.loadF KVServer.S "mu" "ks");;
    #().

Definition MakeKVServer: val :=
  rec: "MakeKVServer" <> :=
    let: "ks" := struct.alloc KVServer.S (zero_val (struct.t KVServer.S)) in
    struct.storeF KVServer.S "mu" "ks" (lock.new #());;
    struct.storeF KVServer.S "kvs" "ks" (NewMap uint64T);;
    struct.storeF KVServer.S "sv" "ks" (MakeRPCServer #());;
    "ks".

Definition KVServer__Start: val :=
  rec: "KVServer__Start" "ks" :=
    let: "handlers" := NewMap grove_common.RawRpcFunc in
    MapInsert "handlers" KV_PUT (ConjugateRpcFunc (KVServer__Put "ks"));;
    MapInsert "handlers" KV_GET (ConjugateRpcFunc (KVServer__Get "ks"));;
    grove_ffi.StartRPCServer "handlers".

(* 1_lockserver.go *)

Module LockServer.
  Definition S := struct.decl [
    "mu" :: lockRefT;
    "sv" :: struct.ptrT RPCServer.S;
    "locks" :: mapT boolT
  ].
End LockServer.

Definition LockServer__tryLock_core: val :=
  rec: "LockServer__tryLock_core" "ls" "args" :=
    let: "lockname" := struct.get grove_common.RPCVals.S "U64_1" "args" in
    let: ("locked", <>) := MapGet (struct.loadF LockServer.S "locks" "ls") "lockname" in
    (if: "locked"
    then #0
    else
      MapInsert (struct.loadF LockServer.S "locks" "ls") "lockname" #true;;
      #1).

Definition LockServer__unlock_core: val :=
  rec: "LockServer__unlock_core" "ls" "args" :=
    let: "lockname" := struct.get grove_common.RPCVals.S "U64_1" "args" in
    let: ("locked", <>) := MapGet (struct.loadF LockServer.S "locks" "ls") "lockname" in
    (if: "locked"
    then
      MapInsert (struct.loadF LockServer.S "locks" "ls") "lockname" #false;;
      #1
    else #0).

(* server Lock RPC handler. *)
Definition LockServer__TryLock: val :=
  rec: "LockServer__TryLock" "ls" "req" "reply" :=
    lock.acquire (struct.loadF LockServer.S "mu" "ls");;
    RPCServer__HandleRequest (struct.loadF LockServer.S "sv" "ls") (LockServer__tryLock_core "ls") "req" "reply";;
    lock.release (struct.loadF LockServer.S "mu" "ls");;
    #().

(* server Unlock RPC handler. *)
Definition LockServer__Unlock: val :=
  rec: "LockServer__Unlock" "ls" "req" "reply" :=
    lock.acquire (struct.loadF LockServer.S "mu" "ls");;
    RPCServer__HandleRequest (struct.loadF LockServer.S "sv" "ls") (LockServer__unlock_core "ls") "req" "reply";;
    lock.release (struct.loadF LockServer.S "mu" "ls");;
    #().

Definition MakeLockServer: val :=
  rec: "MakeLockServer" <> :=
    let: "ls" := struct.alloc LockServer.S (zero_val (struct.t LockServer.S)) in
    struct.storeF LockServer.S "mu" "ls" (lock.new #());;
    struct.storeF LockServer.S "locks" "ls" (NewMap boolT);;
    struct.storeF LockServer.S "sv" "ls" (MakeRPCServer #());;
    "ls".

Definition LockServer__Start: val :=
  rec: "LockServer__Start" "ls" :=
    let: "handlers" := NewMap grove_common.RawRpcFunc in
    MapInsert "handlers" "LOCK_TRYLOCK" (ConjugateRpcFunc (LockServer__TryLock "ls"));;
    MapInsert "handlers" "LOCK_UNLOCK" (ConjugateRpcFunc (LockServer__Unlock "ls"));;
    grove_ffi.StartRPCServer "handlers".

(* 3_kvclient.go *)

Module KVClerk.
  Definition S := struct.decl [
    "host" :: stringT;
    "client" :: struct.ptrT RPCClient.S;
    "cid" :: uint64T;
    "seq" :: uint64T
  ].
End KVClerk.

Definition MakeKVClerk: val :=
  rec: "MakeKVClerk" "host" "cid" :=
    let: "ck" := struct.alloc KVClerk.S (zero_val (struct.t KVClerk.S)) in
    struct.storeF KVClerk.S "host" "ck" "host";;
    struct.storeF KVClerk.S "client" "ck" (MakeRPCClient "host" "cid");;
    struct.storeF KVClerk.S "cid" "ck" "cid";;
    struct.storeF KVClerk.S "seq" "ck" #1;;
    "ck".

Definition KVClerk__Put: val :=
  rec: "KVClerk__Put" "ck" "key" "val" :=
    RPCClient__MakeRequest (struct.loadF KVClerk.S "client" "ck") KV_PUT (struct.mk grove_common.RPCVals.S [
      "U64_1" ::= "key";
      "U64_2" ::= "val"
    ]);;
    #().

Definition KVClerk__Get: val :=
  rec: "KVClerk__Get" "ck" "key" :=
    RPCClient__MakeRequest (struct.loadF KVClerk.S "client" "ck") KV_GET (struct.mk grove_common.RPCVals.S [
      "U64_1" ::= "key"
    ]).

(* 3_lockclient.go *)

(* the lockservice Clerk lives in the client
   and maintains a little state. *)
Module Clerk.
  Definition S := struct.decl [
    "host" :: stringT;
    "client" :: struct.ptrT RPCClient.S
  ].
End Clerk.

Definition LOCK_TRYLOCK : expr := #1.

Definition LOCK_UNLOCK : expr := #2.

Definition MakeClerk: val :=
  rec: "MakeClerk" "host" "cid" :=
    let: "ck" := struct.alloc Clerk.S (zero_val (struct.t Clerk.S)) in
    struct.storeF Clerk.S "host" "ck" "host";;
    struct.storeF Clerk.S "client" "ck" (MakeRPCClient "host" "cid");;
    "ck".

Definition Clerk__TryLock: val :=
  rec: "Clerk__TryLock" "ck" "lockname" :=
    RPCClient__MakeRequest (struct.loadF Clerk.S "client" "ck") LOCK_TRYLOCK (struct.mk grove_common.RPCVals.S [
      "U64_1" ::= "lockname"
    ]) ≠ #0.

(* ask the lock service to unlock a lock.
   returns true if the lock was previously held,
   false otherwise. *)
Definition Clerk__Unlock: val :=
  rec: "Clerk__Unlock" "ck" "lockname" :=
    RPCClient__MakeRequest (struct.loadF Clerk.S "client" "ck") LOCK_UNLOCK (struct.mk grove_common.RPCVals.S [
      "U64_1" ::= "lockname"
    ]) ≠ #0.

(* Spins until we have the lock *)
Definition Clerk__Lock: val :=
  rec: "Clerk__Lock" "ck" "lockname" :=
    Skip;;
    (for: (λ: <>, #true); (λ: <>, Skip) := λ: <>,
      (if: Clerk__TryLock "ck" "lockname"
      then Break
      else Continue));;
    #true.

(* 4_bank.go *)

Module Bank.
  Definition S := struct.decl [
    "ls" :: stringT;
    "ks" :: stringT
  ].
End Bank.

Module BankClerk.
  Definition S := struct.decl [
    "lck" :: struct.ptrT Clerk.S;
    "kvck" :: struct.ptrT KVClerk.S;
    "acc1" :: uint64T;
    "acc2" :: uint64T
  ].
End BankClerk.

Definition acquire_two: val :=
  rec: "acquire_two" "lck" "l1" "l2" :=
    (if: "l1" < "l2"
    then
      Clerk__Lock "lck" "l1";;
      Clerk__Lock "lck" "l2"
    else
      Clerk__Lock "lck" "l2";;
      Clerk__Lock "lck" "l1");;
    #().

Definition release_two: val :=
  rec: "release_two" "lck" "l1" "l2" :=
    (if: "l1" < "l2"
    then
      Clerk__Unlock "lck" "l2";;
      Clerk__Unlock "lck" "l1"
    else
      Clerk__Unlock "lck" "l1";;
      Clerk__Unlock "lck" "l2");;
    #().

(* Requires that the account numbers are smaller than num_accounts
   If account balance in acc_from is at least amount, transfer amount to acc_to *)
Definition BankClerk__transfer_internal: val :=
  rec: "BankClerk__transfer_internal" "bck" "acc_from" "acc_to" "amount" :=
    acquire_two (struct.loadF BankClerk.S "lck" "bck") "acc_from" "acc_to";;
    let: "old_amount" := KVClerk__Get (struct.loadF BankClerk.S "kvck" "bck") "acc_from" in
    (if: "old_amount" ≥ "amount"
    then
      KVClerk__Put (struct.loadF BankClerk.S "kvck" "bck") "acc_from" ("old_amount" - "amount");;
      KVClerk__Put (struct.loadF BankClerk.S "kvck" "bck") "acc_to" (KVClerk__Get (struct.loadF BankClerk.S "kvck" "bck") "acc_to" + "amount");;
      #()
    else #());;
    release_two (struct.loadF BankClerk.S "lck" "bck") "acc_from" "acc_to".

Definition BankClerk__SimpleTransfer: val :=
  rec: "BankClerk__SimpleTransfer" "bck" "amount" :=
    BankClerk__transfer_internal "bck" (struct.loadF BankClerk.S "acc1" "bck") (struct.loadF BankClerk.S "acc2" "bck") "amount".

(* If account balance in acc_from is at least amount, transfer amount to acc_to *)
Definition BankClerk__SimpleAudit: val :=
  rec: "BankClerk__SimpleAudit" "bck" :=
    acquire_two (struct.loadF BankClerk.S "lck" "bck") (struct.loadF BankClerk.S "acc1" "bck") (struct.loadF BankClerk.S "acc2" "bck");;
    let: "sum" := KVClerk__Get (struct.loadF BankClerk.S "kvck" "bck") (struct.loadF BankClerk.S "acc1" "bck") + KVClerk__Get (struct.loadF BankClerk.S "kvck" "bck") (struct.loadF BankClerk.S "acc2" "bck") in
    release_two (struct.loadF BankClerk.S "lck" "bck") (struct.loadF BankClerk.S "acc1" "bck") (struct.loadF BankClerk.S "acc2" "bck");;
    "sum".

Definition MakeBank: val :=
  rec: "MakeBank" "lsid" "ksid" :=
    struct.mk Bank.S [
      "ls" ::= "lsid";
      "ks" ::= "ksid"
    ].

Definition MakeBankClerk: val :=
  rec: "MakeBankClerk" "b" "acc1" "acc2" "cid" :=
    let: "bck" := struct.alloc BankClerk.S (zero_val (struct.t BankClerk.S)) in
    struct.storeF BankClerk.S "lck" "bck" (MakeClerk (struct.get Bank.S "ls" "b") "cid");;
    struct.storeF BankClerk.S "kvck" "bck" (MakeKVClerk (struct.get Bank.S "ks" "b") "cid");;
    struct.storeF BankClerk.S "acc1" "bck" "acc1";;
    struct.storeF BankClerk.S "acc2" "bck" "acc2";;
    "bck".

(* 4_incrserver.go *)

Definition IS_INCR : expr := #1.

Module IncrServer.
  Definition S := struct.decl [
    "mu" :: lockRefT;
    "sv" :: struct.ptrT RPCServer.S;
    "kvserver" :: struct.ptrT KVServer.S;
    "kck" :: struct.ptrT KVClerk.S
  ].
End IncrServer.

Definition IncrServer__increment_core_unsafe: val :=
  rec: "IncrServer__increment_core_unsafe" "is" "seq" "args" :=
    let: "key" := struct.get grove_common.RPCVals.S "U64_1" "args" in
    let: "oldv" := ref (zero_val uint64T) in
    "oldv" <-[uint64T] KVClerk__Get (struct.loadF IncrServer.S "kck" "is") "key";;
    KVClerk__Put (struct.loadF IncrServer.S "kck" "is") "key" (![uint64T] "oldv" + #1);;
    #0.

(* crash-safely increment counter and return the new value

   Idea is this:
   In the unsafe version, we ought to have the quadruples

   { key [kv]-> a }
    A
   { key [kv]-> _ \ast v = a }
   { key [kv]-> a }

   { key [kv]-> _ \ast v = a \ast durable_oldv = a }
    B
   { key [kv]-> (a + 1) }
   { key [kv]-> _ \ast durable_oldv = a }

   By adding code between A and B that makes durable_oldv = v, we can ensure
   that rerunning the function will result in B starting with the correct
   durable_oldv.
   TODO: test this
   Probably won't try proving this version correct (first). *)
Definition IncrServer__increment_core: val :=
  rec: "IncrServer__increment_core" "is" "seq" "args" :=
    let: "key" := struct.get grove_common.RPCVals.S "U64_1" "args" in
    let: "oldv" := ref (zero_val uint64T) in
    let: "filename" := #(str"incr_request_") + grove_ffi.U64ToString "seq" + #(str"_oldv") in
    let: "content" := grove_ffi.Read "filename" in
    (if: slice.len "content" > #0
    then "oldv" <-[uint64T] marshal.Dec__GetInt (marshal.NewDec (grove_ffi.Read "filename"))
    else
      "oldv" <-[uint64T] KVClerk__Get (struct.loadF IncrServer.S "kck" "is") "key";;
      let: "enc" := marshal.NewEnc #8 in
      marshal.Enc__PutInt "enc" (![uint64T] "oldv");;
      grove_ffi.Write "filename" (marshal.Enc__Finish "enc"));;
    KVClerk__Put (struct.loadF IncrServer.S "kck" "is") "key" (![uint64T] "oldv" + #1);;
    #0.

(* For now, there is only one kv server in the whole world *)
Definition WriteDurableIncrServer: val :=
  rec: "WriteDurableIncrServer" "ks" :=
    #().

Definition IncrServer__Increment: val :=
  rec: "IncrServer__Increment" "is" "req" "reply" :=
    lock.acquire (struct.loadF IncrServer.S "mu" "is");;
    (if: CheckReplyTable (struct.loadF RPCServer.S "lastSeq" (struct.loadF IncrServer.S "sv" "is")) (struct.loadF RPCServer.S "lastReply" (struct.loadF IncrServer.S "sv" "is")) (struct.loadF grove_common.RPCRequest.S "CID" "req") (struct.loadF grove_common.RPCRequest.S "Seq" "req") "reply"
    then #()
    else
      struct.storeF grove_common.RPCReply.S "Ret" "reply" (IncrServer__increment_core "is" (struct.loadF grove_common.RPCRequest.S "Seq" "req") (struct.loadF grove_common.RPCRequest.S "Args" "req"));;
      MapInsert (struct.loadF RPCServer.S "lastReply" (struct.loadF IncrServer.S "sv" "is")) (struct.loadF grove_common.RPCRequest.S "CID" "req") (struct.loadF grove_common.RPCReply.S "Ret" "reply");;
      WriteDurableIncrServer "is");;
    lock.release (struct.loadF IncrServer.S "mu" "is");;
    #false.

Definition ReadDurableIncrServer: val :=
  rec: "ReadDurableIncrServer" <> :=
    slice.nil.

Definition MakeIncrServer: val :=
  rec: "MakeIncrServer" "kvserver" :=
    let: "is_old" := ReadDurableIncrServer #() in
    (if: "is_old" ≠ #null
    then "is_old"
    else
      let: "is" := struct.alloc IncrServer.S (zero_val (struct.t IncrServer.S)) in
      struct.storeF IncrServer.S "mu" "is" (lock.new #());;
      struct.storeF IncrServer.S "sv" "is" (MakeRPCServer #());;
      struct.storeF IncrServer.S "kvserver" "is" "kvserver";;
      "is").

(* 5_incrclient.go *)

Module IncrClerk.
  Definition S := struct.decl [
    "host" :: stringT;
    "client" :: struct.ptrT RPCClient.S;
    "cid" :: uint64T;
    "seq" :: uint64T
  ].
End IncrClerk.

Definition MakeIncrClerk: val :=
  rec: "MakeIncrClerk" "host" "cid" :=
    let: "ck" := struct.alloc IncrClerk.S (zero_val (struct.t IncrClerk.S)) in
    struct.storeF IncrClerk.S "host" "ck" "host";;
    struct.storeF IncrClerk.S "client" "ck" (MakeRPCClient "host" "cid");;
    "ck".

Definition IncrClerk__Increment: val :=
  rec: "IncrClerk__Increment" "ck" "key" :=
    RPCClient__MakeRequest (struct.loadF IncrClerk.S "client" "ck") IS_INCR (struct.mk grove_common.RPCVals.S [
      "U64_1" ::= "key"
    ]);;
    #().

(* 6_incrproxyserver.go *)

Module IncrProxyServer.
  Definition S := struct.decl [
    "sv" :: struct.ptrT RPCServer.S;
    "incrserver" :: stringT;
    "ick" :: struct.ptrT IncrClerk.S;
    "lastCID" :: uint64T
  ].
End IncrProxyServer.

Definition IncrProxyServer__proxy_increment_core_unsafe: val :=
  rec: "IncrProxyServer__proxy_increment_core_unsafe" "is" "seq" "args" :=
    let: "key" := struct.get grove_common.RPCVals.S "U64_1" "args" in
    IncrClerk__Increment (struct.loadF IncrProxyServer.S "ick" "is") "key";;
    #0.

(* Common code for RPC clients: tracking of CID and next sequence number. *)
Module ShortTermIncrClerk.
  Definition S := struct.decl [
    "cid" :: uint64T;
    "seq" :: uint64T;
    "req" :: struct.t grove_common.RPCRequest.S;
    "incrserver" :: stringT
  ].
End ShortTermIncrClerk.

Definition ShortTermIncrClerk__PrepareRequest: val :=
  rec: "ShortTermIncrClerk__PrepareRequest" "ck" "args" :=
    overflow_guard_incr (struct.loadF ShortTermIncrClerk.S "seq" "ck");;
    struct.storeF ShortTermIncrClerk.S "req" "ck" (struct.mk grove_common.RPCRequest.S [
      "Args" ::= "args";
      "CID" ::= struct.loadF ShortTermIncrClerk.S "cid" "ck";
      "Seq" ::= struct.loadF ShortTermIncrClerk.S "seq" "ck"
    ]);;
    struct.storeF ShortTermIncrClerk.S "seq" "ck" (struct.loadF ShortTermIncrClerk.S "seq" "ck" + #1).

Definition ShortTermIncrClerk__MakePreparedRequest: val :=
  rec: "ShortTermIncrClerk__MakePreparedRequest" "ck" :=
    let: "errb" := ref_to boolT #false in
    let: "reply" := struct.alloc grove_common.RPCReply.S (zero_val (struct.t grove_common.RPCReply.S)) in
    let: "cl" := grove_ffi.MakeRPCClient (struct.loadF ShortTermIncrClerk.S "incrserver" "ck") in
    Skip;;
    (for: (λ: <>, #true); (λ: <>, Skip) := λ: <>,
      "errb" <-[boolT] RemoteProcedureCall2 "cl" IS_INCR (struct.fieldRef ShortTermIncrClerk.S "req" "ck") "reply";;
      (if: (![boolT] "errb" = #false)
      then Break
      else Continue));;
    struct.loadF grove_common.RPCReply.S "Ret" "reply".

Definition DecodeShortTermIncrClerk: val :=
  rec: "DecodeShortTermIncrClerk" "is" "content" :=
    let: "d" := marshal.NewDec "content" in
    let: "ck" := struct.alloc ShortTermIncrClerk.S (zero_val (struct.t ShortTermIncrClerk.S)) in
    struct.storeF ShortTermIncrClerk.S "incrserver" "ck" "is";;
    struct.storeF ShortTermIncrClerk.S "cid" "ck" (marshal.Dec__GetInt "d");;
    struct.storeF ShortTermIncrClerk.S "seq" "ck" (marshal.Dec__GetInt "d");;
    struct.storeF grove_common.RPCRequest.S "CID" (struct.fieldRef ShortTermIncrClerk.S "req" "ck") (struct.loadF ShortTermIncrClerk.S "cid" "ck");;
    struct.storeF grove_common.RPCRequest.S "Seq" (struct.fieldRef ShortTermIncrClerk.S "req" "ck") (struct.loadF ShortTermIncrClerk.S "seq" "ck" - #1);;
    struct.storeF grove_common.RPCVals.S "U64_1" (struct.fieldRef grove_common.RPCRequest.S "Args" (struct.fieldRef ShortTermIncrClerk.S "req" "ck")) (marshal.Dec__GetInt "d");;
    struct.storeF grove_common.RPCVals.S "U64_2" (struct.fieldRef grove_common.RPCRequest.S "Args" (struct.fieldRef ShortTermIncrClerk.S "req" "ck")) (marshal.Dec__GetInt "d");;
    "ck".

Definition EncodeShortTermIncrClerk: val :=
  rec: "EncodeShortTermIncrClerk" "ck" :=
    let: "e" := marshal.NewEnc #32 in
    marshal.Enc__PutInt "e" (struct.loadF ShortTermIncrClerk.S "cid" "ck");;
    marshal.Enc__PutInt "e" (struct.loadF ShortTermIncrClerk.S "seq" "ck");;
    marshal.Enc__PutInt "e" (struct.get grove_common.RPCVals.S "U64_1" (struct.get grove_common.RPCRequest.S "Args" (struct.loadF ShortTermIncrClerk.S "req" "ck")));;
    marshal.Enc__PutInt "e" (struct.get grove_common.RPCVals.S "U64_2" (struct.get grove_common.RPCRequest.S "Args" (struct.loadF ShortTermIncrClerk.S "req" "ck")));;
    marshal.Enc__Finish "e".

Definition IncrProxyServer__MakeFreshIncrClerk: val :=
  rec: "IncrProxyServer__MakeFreshIncrClerk" "is" :=
    let: "cid" := struct.loadF IncrProxyServer.S "lastCID" "is" in
    overflow_guard_incr (struct.loadF IncrProxyServer.S "lastCID" "is");;
    struct.storeF IncrProxyServer.S "lastCID" "is" (struct.loadF IncrProxyServer.S "lastCID" "is" + #1);;
    let: "e" := marshal.NewEnc #8 in
    marshal.Enc__PutInt "e" (struct.loadF IncrProxyServer.S "lastCID" "is");;
    grove_ffi.Write #(str"lastCID") (marshal.Enc__Finish "e");;
    let: "ck_ptr" := struct.alloc ShortTermIncrClerk.S (zero_val (struct.t ShortTermIncrClerk.S)) in
    struct.storeF ShortTermIncrClerk.S "cid" "ck_ptr" "cid";;
    struct.storeF ShortTermIncrClerk.S "seq" "ck_ptr" #1;;
    struct.storeF ShortTermIncrClerk.S "incrserver" "ck_ptr" (struct.loadF IncrProxyServer.S "incrserver" "is");;
    "ck_ptr".

Definition IncrProxyServer__proxy_increment_core: val :=
  rec: "IncrProxyServer__proxy_increment_core" "is" "seq" "args" :=
    let: "filename" := #(str"procy_incr_request_") + grove_ffi.U64ToString "seq" in
    let: "ck" := ref (zero_val (refT (struct.t ShortTermIncrClerk.S))) in
    let: "content" := grove_ffi.Read "filename" in
    (if: slice.len "content" > #0
    then "ck" <-[refT (struct.t ShortTermIncrClerk.S)] DecodeShortTermIncrClerk (struct.loadF IncrProxyServer.S "incrserver" "is") "content"
    else
      "ck" <-[refT (struct.t ShortTermIncrClerk.S)] IncrProxyServer__MakeFreshIncrClerk "is";;
      ShortTermIncrClerk__PrepareRequest (![refT (struct.t ShortTermIncrClerk.S)] "ck") "args";;
      let: "content" := EncodeShortTermIncrClerk (![refT (struct.t ShortTermIncrClerk.S)] "ck") in
      grove_ffi.Write "filename" "content");;
    ShortTermIncrClerk__MakePreparedRequest (![refT (struct.t ShortTermIncrClerk.S)] "ck");;
    #0.
